/*++
 Copyright (c) 2008, Intel Corporation
 All rights reserved. 

 This UEFI SCT ("Software") is furnished under license and may only be used or copied
 in accordance with the terms of that license UEFI Contributor Agreement.
 No license, express or implied, by estoppel or otherwise, to any intellectual property
 rights is granted by this document. The Software is subject to change without notice,
 and should not be construed as a commitment by Intel Corporation to market, license,
 sell or support any product or technology. Unless otherwise provided for in the license
 under which this Software is provided, the Software is provided AS IS, with no warranties
 of any kind, express or implied. Except as expressly permitted by the Software license,
 neither Intel Corporation nor its suppliers assumes any responsibility or liability
 for any errors or inaccuracies that may appear herein.
--*/
/*++

Module Name:

  lock.c

Abstract:

  Implements FLOCK

--*/

#include "lib.h"


VOID
InitializeLock (
  IN OUT FLOCK    *Lock,
  IN EFI_TPL      Priority
  )
/*++

Routine Description:

  Initialize a basic mutual exclusion lock.   Each lock
  provides mutual exclusion access at it's task priority
  level.  Since there is no-premption (at any TPL) or
  multiprocessor support, acquiring the lock only consists
  of raising to the locks TPL.

  Note on a debug build the lock is acquired and released
  to help ensure proper usage.

Arguments:

  Lock        - The FLOCK structure to initialize

  Priority    - The task priority level of the lock


Returns:

  An initialized F Lock structure.

--*/
{
  Lock->Tpl = Priority;
  Lock->OwnerTpl = 0;
  Lock->Lock = 0;
}


VOID
AcquireLock (
  IN FLOCK    *Lock
  )
/*++

Routine Description:

  Raising to the task priority level of the mutual exclusion
  lock, and then acquires ownership of the lock.

Arguments:

  Lock        - The lock to acquire

Returns:

  Lock owned

--*/
{
  if (tBS) {
    if (tBS->RaiseTPL != NULL) {
      Lock->OwnerTpl = tBS->RaiseTPL(Lock->Tpl);
    }
  } else {
    if (LibRuntimeRaiseTPL != NULL) {
      Lock->OwnerTpl = LibRuntimeRaiseTPL(Lock->Tpl);
    }
  }
  Lock->Lock += 1;
  ASSERT (Lock->Lock == 1);
}


VOID
ReleaseLock (
  IN FLOCK    *Lock
  )
/*++

Routine Description:

  Releases ownership of the mutual exclusion lock, and
  restores the previous task priority level.

Arguments:

  Lock        - The lock to release

Returns:

  Lock unowned

--*/
{
  EFI_TPL     Tpl;

  Tpl = Lock->OwnerTpl;
  ASSERT(Lock->Lock == 1);
  Lock->Lock -= 1;
  if (tBS) {
    if (tBS->RestoreTPL != NULL) {
      tBS->RestoreTPL (Tpl);
    }
  } else {
    if (LibRuntimeRestoreTPL != NULL) {
      LibRuntimeRestoreTPL(Tpl);
    }
  }
}
